#include "treelib.def"
!=========================================
!
! coreregions.F
!
! PURPOSE: finds multiple cores in one volume, marks
!          them with a mask and stores the total mask.
!
!          core finding: start at density maximum and search
!                        for coherent regions above density or
!                        below temperature threshold.
!          core marking: core cells are marked by current number
!                        of core (integer mask).
!          core output : integer array, same size as input grid,
!                        with cells = 0 except for cores. 
!
!=========================================

program coreregions

  use treelib ! module treelib.F, containing tree-functions

  implicit none

  integer nx, ny, nz, imax, jmax, kmax, i, j, k, ilo, ihi, jlo, jhi, klo, khi
  integer logunit, dmpunit, inpunit, idebug, count
  integer dmp, icore, ndim, iter
  integer maxlevel, icrit, idone, imode, ntreesize, nmincore

  logical acceptnode

#ifdef SINGLE
  real   acceptthresh, maxdens, timec, mintemp, tiny, huge, perc
  parameter (tiny=-1e30, huge=1e30)
#else
  double precision   acceptthresh, maxdens, timec, mintemp, tiny, huge, perc
  parameter (tiny=-1d30, huge=1d30)
#endif

#ifdef SINGLE
  real,allocatable :: dens(:,:,:)
  real,allocatable :: prss(:,:,:)
  real,allocatable :: temp(:,:,:)
#else
  double precision,allocatable :: dens(:,:,:)
  double precision,allocatable :: prss(:,:,:)
  double precision,allocatable :: temp(:,:,:)
#endif

  integer,allocatable :: core(:,:,:)
  logical,allocatable :: free(:,:,:)

  type(treenode),pointer :: proot

  character(12) filename
  character(4)  filestem

  external acceptnode

!---------------------------------------

  inpunit = 1
  dmpunit = 4
  
  open(unit=inpunit,file="coreregions.inp",status="unknown")
  read(inpunit,*) logunit   ! 2: file, 6: screen
  read(inpunit,*) idebug    ! level of debugging messages
  read(inpunit,*) nx        ! nx, ny, nz domain size. If 2D, use nz=1
  read(inpunit,*) ny
  read(inpunit,*) nz
  read(inpunit,*) ilo, ihi  ! lower and upper indices constraining search volume
  read(inpunit,*) jlo, jhi
  read(inpunit,*) klo, khi
  read(inpunit,*) dmp        ! dump number
  read(inpunit,*) icore      ! not used
  read(inpunit,*) maxlevel   ! currently not used.
  read(inpunit,*) acceptthresh  ! see acceptnode
  read(inpunit,*) perc       ! additional acceptance crit: only down to perc of maxdens
  read(inpunit,*) nmincore   ! minimum cell number for core to be accepted
  read(inpunit,*) imode      ! 1: search dens. 2: search temp
  read(inpunit,*) icrit      ! criterion number for acceptnode. Use 1 only.
  if (imode .eq. 3) read(inpunit,*) filestem ! if not dens or temp
  close(inpunit)

  if (logunit .eq. 2) then 
    open(unit=logunit,file="coreregions.log",status="unknown")
  endif
    
  allocate(dens(nx,ny,nz))
  if (imode .eq. 2) then 
    allocate(prss(nx,ny,nz))
  endif
  allocate(free(nx,ny,nz))
  allocate(core(nx,ny,nz))

  free(:,:,:) = .true. ! meaning: not occupied
  core(:,:,:) = 0

! read the domain
  if (imode .eq. 3) then
    write(filename,'(a4,i4.4,a4)') filestem, dmp,".f7u"
  else
    write(filename,'(a4,i4.4,a4)') "dens", dmp,".f7u"
  endif
  call proteusread3d(nx,ny,nz,filename,dmp,iter,timec,dens,dmpunit,logunit)
  if (imode .eq. 2) then
    write(filename,'(a4,i4.4,a4)') "prss", dmp,".f7u"
    call proteusread3d(nx,ny,nz,filename,dmp,iter,timec,prss,dmpunit,logunit)
    dens(:,:,:) = prss(:,:,:) / dens(:,:,:)
  endif

  idone  = 0  ! assume there are regions with dens > rhothresh
  count  = 1  ! core counter

  do while (idone .eq. 0)
    if ((imode .eq. 1) .or. (imode .eq. 3)) then
      maxdens = tiny
      do k = klo,khi
        do j = jlo,jhi
          do i = ilo,ihi
            if (core(i,j,k) .eq. 0) then ! only check cells wich haven't been assigned yet
              if ((maxdens .lt. dens(i,j,k)).and.(dens(i,j,k) .ge. acceptthresh)) then
                maxdens = dens(i,j,k) 
                imax   = i
                jmax   = j
                kmax   = k
              endif
            endif
          enddo
        enddo
      enddo
      if (maxdens .eq. tiny) then
        write(logunit,'("COREREGIONS: count ",i4,": no maximum found")') count
        idone = 1
      else
        write(logunit,'("COREREGIONS: count ",i4,": max(dens(",i4,i4,i4,")) = ",e13.5)') count,imax,jmax,kmax,maxdens
!       this is not exactly clean... It's the allocation of the root-node.
        call newnode(proot, maxdens, imax, jmax, kmax, 0, null(), 0, logunit, idebug)
        free(:,:,:) = .true.
!       build the tree, starting at proot and maxdens.
        ntreesize = 0
        call buildtree(proot, nx, ny, nz, free, dens, acceptthresh, maxdens, acceptnode, icrit, ntreesize, logunit, idebug)
        write(logunit,'("COREREGIONS: count ",i4,": building tree finished: ",i8," elements.")') count, ntreesize
        where((.not. free).and.(dens .ge. perc*maxdens).and.(ntreesize .ge. nmincore))
          core(:,:,:) = count
        endwhere
        where((.not. free).and.((dens .lt. perc*maxdens).or.(ntreesize .lt. nmincore)))
          core(:,:,:) = -1
        endwhere
!       deallocate the tree (which can be rather large)
        call deletetree(proot, logunit, idebug)
        write(logunit,'("COREREGIONS: count ",i4,": tree deleted.")') count
        count = count + 1
      endif
    endif

!   temperature branch
    if (imode .eq. 2) then
      mintemp = huge
      do k = klo,khi
        do j = jlo,jhi
          do i = ilo,ihi
            if (core(i,j,k) .eq. 0) then ! only check cells wich haven't been assigned yet
              if ((mintemp .gt. dens(i,j,k)).and.(dens(i,j,k) .le. acceptthresh)) then
                mintemp= dens(i,j,k)
                imax   = i
                jmax   = j
                kmax   = k
              endif
            endif
          enddo
        enddo
      enddo
      if (mintemp .eq. huge) then
        write(logunit,'("COREREGIONS: count ",i4,": no minimum found")') count
        idone = 1
      else
        write(logunit,'("COREREGIONS: count ",i4,": min(temp(",i4,i4,i4,")) = ",e13.5)') count,imax,jmax,kmax,mintemp
!       this is not exactly clean... It's the allocation of the root-node.
        call newnode(proot, mintemp, imax, jmax, kmax, 0, null(), 0, logunit, idebug)
        free(:,:,:) = .true.
!       build the tree, starting at proot and maxdens.
        ntreesize = 0
        call buildtree(proot, nx, ny, nz, free, dens, mintemp, acceptthresh, acceptnode, icrit, ntreesize, logunit, idebug)
        write(logunit,'("COREREGIONS: count ",i4,": building tree finished: ",i8," elements.")') count, ntreesize
        where((.not. free) .and. (ntreesize .ge. nmincore))
          core(:,:,:) = count
        endwhere
        where((.not. free) .and. (ntreesize .lt. nmincore))
          core(:,:,:) = -1
        endwhere
!       deallocate the tree (which can be rather large)
        call deletetree(proot, logunit, idebug)
        write(logunit,'("COREREGIONS: level ",i4,": tree deleted.")') count
        count = count + 1
      endif ! if (mintemp .eq. huge)
    endif ! if (imode .eq. 2)
  enddo ! do while (idone .eq. 0)
  
  if (count .gt. 1) then ! cores have been found
    if (imode .eq. 1) write(filename,'(a4,i4.4,a4)') "cord",dmp,".f7u"
    if (imode .eq. 2) write(filename,'(a4,i4.4,a4)') "cort",dmp,".f7u"
    if (imode .eq. 3) write(filename,'(a1,a3,i4.4,a4)') "c",filestem(2:4),dmp,".f7u"
!    if (nz .eq. 1) then 
!#ifdef SINGLE
!      call proteuswrite2d(nx,ny,filename,dmp,iter,timec,real(core(:,:,1)),dmpunit,logunit)
!#else
!      call proteuswrite2d(nx,ny,filename,dmp,iter,timec,dble(core(:,:,1)),dmpunit,logunit)
!#endif
!    else
#ifdef SINGLE
      call proteuswrite3d(nx,ny,nz,filename,dmp,iter,timec,real(core(:,:,:)),dmpunit,logunit)
#else
      call proteuswrite3d(nx,ny,nz,filename,dmp,iter,timec,dble(core(:,:,:)),dmpunit,logunit)
#endif
!    endif
  endif

  deallocate(dens)
  if (imode .eq. 2) deallocate(prss)
  deallocate(free)
  deallocate(core)

  if (logunit .eq. 2) then
    close(logunit)
  endif

!====================================================

end
