#include "treelib.def"
!============================================
!
! treelib.F
!
! PURPOSE: library routines for coretree
!          A treenode can have six branches at most,
!          one for each coordinate direction.
!
!============================================

module treelib

  use datatypes  ! f90-lingo. module datatypes contains the type declaration for treenode
 
  contains       ! again f90-lingo. No idea why that is necessary.

!--------------------------------------------
! allocates new node with given parameters.
! returns pointer of type treenode.
!--------------------------------------------
subroutine newnode(pnew, rho, i, j, k, mother, pcaller, level, logunit, idebug)

  implicit none
 
  integer idebug  ! debugging level
  integer logunit !
  integer level   ! node level
  integer i, j, k ! cell coordinates
  integer mother  ! denoting in which direction mother node is located (1..6)

#ifdef SINGLE
  real             rho     ! density at (i,j,k)
#else
  double precision rho     ! density at (i,j,k)
#endif

  type(treenode),pointer :: pcaller ! pointer to mother node
  type(treenode),pointer :: pnew    ! result returned by newnode

  allocate(pnew)
  pnew%pnodeip => null()
  pnew%pnodejp => null()
  pnew%pnodekp => null()
  pnew%pnodeim => null()
  pnew%pnodejm => null()
  pnew%pnodekm => null()
  pnew%pmother => pcaller
  pnew%imother =  mother 
  pnew%rho     =  rho
  pnew%i       =  i
  pnew%j       =  j
  pnew%k       =  k
  pnew%level   =  level
 
  return
end subroutine newnode

!--------------------------------------------
! constructs tree from data array rho recursively, starting
! at pnode. For pnode = root, call with level = 0
! The acceptance criterion is hidden in acceptnode.F 
! Returns .true. if cell should be accepted.
!--------------------------------------------
recursive subroutine buildtree(pnode, nx, ny, nz, mask, rho, rhothresh, rhomax, &
                               acceptnode, icrit, ntreesize, logunit, idebug)

  implicit none

  integer idebug                    ! debugging level
  integer logunit                   ! 
  integer nx, ny, nz                ! size of rho
  integer icrit                     ! criterion number for acceptnode (see there)
  integer ntreesize                 ! number of elements
  logical free                      ! temporary for existing node
  logical indomain                  ! checking if array is within bounds
  logical acceptnode                ! function for accepting new node. Returns true if accepted.
#ifdef SINGLE
  real             trho                      ! temporary local rho of prospective new node
  real             crho                      ! current rho
  real             rhothresh                 ! density threshold
  real             rhomax                    ! maxval(rho)
  real             rho (nx, ny, nz)          ! density array
#else
  double precision trho                      ! temporary local rho of prospective new node
  double precision crho                      ! current rho
  double precision rhothresh                 ! density threshold
  double precision rhomax                    ! maxval(rho)
  double precision rho (nx, ny, nz)          ! density array
#endif
  logical mask(nx, ny, nz)          ! mask array checking whether we have coordinate.
  type(treenode),pointer :: pnode   ! "root" node

  external acceptnode

!----------------------------

  mask(pnode%i, pnode%j, pnode%k) = .false.
  crho                            = rho(pnode%i, pnode%j, pnode%k)
  ntreesize                       = ntreesize + 1

!  if (idebug .gt. 1) then
!    write(logunit,'("BUILDTREE: level,i,j,k = ",i5,i4,i4,i4)') pnode%level,pnode%i, pnode%j, pnode%k
!  endif

  if (pnode%imother .ne. 1) then    ! we are not coming from i+1
    indomain = (pnode%i + 1 .le. nx)
    if (indomain) then
      trho     = rho (pnode%i + 1, pnode%j, pnode%k)
      free     = mask(pnode%i + 1, pnode%j, pnode%k)
      if (free .and. acceptnode(crho, trho, rhothresh, rhomax, icrit)) then
        call newnode  (pnode%pnodeip, trho, pnode%i + 1, pnode%j, pnode%k, 4, pnode, pnode%level+1, logunit, idebug)
        call buildtree(pnode%pnodeip, nx, ny, nz, mask, rho, rhothresh, rhomax, acceptnode, &
                       icrit, ntreesize, logunit, idebug)
      endif
    endif
  endif
  if (pnode%imother .ne. 2) then    ! we are not coming from j+1
    indomain = (pnode%j + 1 .le. ny)
    if (indomain) then
      trho     = rho (pnode%i, pnode%j + 1, pnode%k)
      free     = mask(pnode%i, pnode%j + 1, pnode%k)
      if (free .and. acceptnode(crho, trho, rhothresh, rhomax, icrit)) then
        call newnode  (pnode%pnodejp, trho, pnode%i, pnode%j + 1, pnode%k, 5, pnode, pnode%level+1, logunit, idebug)
        call buildtree(pnode%pnodejp, nx, ny, nz, mask, rho, rhothresh, rhomax, acceptnode, &
                       icrit, ntreesize, logunit, idebug)
      endif
    endif
  endif
  if (pnode%imother .ne. 3) then    ! we are not coming from k+1
    indomain = (pnode%k + 1 .le. nz)
    if (indomain) then
      trho     = rho (pnode%i, pnode%j, pnode%k + 1)
      free     = mask(pnode%i, pnode%j, pnode%k + 1)
      if (free .and. acceptnode(crho, trho, rhothresh, rhomax, icrit)) then
        call newnode  (pnode%pnodekp, trho, pnode%i, pnode%j, pnode%k + 1, 6, pnode, pnode%level+1, logunit, idebug)
        call buildtree(pnode%pnodekp, nx, ny, nz, mask, rho, rhothresh, rhomax, acceptnode, &
                       icrit, ntreesize, logunit, idebug)
      endif
    endif
  endif
  if (pnode%imother .ne. 4) then    ! we are not coming from i-1
    indomain = (pnode%i - 1 .ge. 1)
    if (indomain) then
      trho     = rho (pnode%i - 1, pnode%j, pnode%k)
      free     = mask(pnode%i - 1, pnode%j, pnode%k)
      if (free .and. acceptnode(crho, trho, rhothresh, rhomax, icrit)) then
        call newnode  (pnode%pnodeim, trho, pnode%i - 1, pnode%j, pnode%k, 1, pnode, pnode%level+1, logunit, idebug)
        call buildtree(pnode%pnodeim, nx, ny, nz, mask, rho, rhothresh, rhomax, acceptnode, &
                       icrit, ntreesize, logunit, idebug)
      endif
    endif
  endif
  if (pnode%imother .ne. 5) then    ! we are not coming from j-1
    indomain = (pnode%j - 1 .ge. 1)
    if (indomain) then
      trho     = rho (pnode%i, pnode%j - 1, pnode%k)
      free     = mask(pnode%i, pnode%j - 1, pnode%k)
      if (free .and. acceptnode(crho, trho, rhothresh, rhomax, icrit)) then
        call newnode  (pnode%pnodejm, trho, pnode%i, pnode%j - 1, pnode%k, 2, pnode, pnode%level+1, logunit, idebug)
        call buildtree(pnode%pnodejm, nx, ny, nz, mask, rho, rhothresh, rhomax, acceptnode, &
                       icrit, ntreesize, logunit, idebug)
      endif
    endif
  endif
  if (pnode%imother .ne. 6) then    ! we are not coming from k+1
    indomain = (pnode%k - 1 .ge. 1)
    if (indomain) then
      trho     = rho (pnode%i, pnode%j, pnode%k - 1)
      free     = mask(pnode%i, pnode%j, pnode%k - 1)
      if (free .and. acceptnode(crho, trho, rhothresh, rhomax, icrit)) then
        call newnode  (pnode%pnodekm, trho, pnode%i, pnode%j, pnode%k - 1, 3, pnode, pnode%level+1, logunit, idebug)
        call buildtree(pnode%pnodekm, nx, ny, nz, mask, rho, rhothresh, rhomax, acceptnode, &
                       icrit, ntreesize, logunit, idebug)
      endif
    endif
  endif

  return
end subroutine buildtree

!--------------------------------------------
! walks through tree and does something.
! probably printing
!--------------------------------------------
recursive subroutine walktree(pnode, logunit, dmpunit, idebug)

  implicit none

  integer logunit, dmpunit, idebug

  type(treenode),pointer :: pnode

!----------------------------

  write(logunit,'("WALKTREE: ",i4,", rho(",i4,",",i4,",",i4,")=",e13.5)') &
       pnode%level, pnode%i, pnode%j, pnode%k, pnode%rho

  if (dmpunit .ne. 0) then
    write(dmpunit) pnode%i, pnode%j, pnode%k, pnode%rho
  endif

  if (pnode%imother .ne. 1) then    ! we are not coming from i+1
    if (associated(pnode%pnodeip)) then
      call walktree(pnode%pnodeip, logunit, dmpunit, idebug)
    endif
  endif
  if (pnode%imother .ne. 2) then    ! we are not coming from j+1
    if (associated(pnode%pnodejp)) then
      call walktree(pnode%pnodejp, logunit, dmpunit, idebug)
    endif
  endif
  if (pnode%imother .ne. 3) then    ! we are not coming from k+1
    if (associated(pnode%pnodekp)) then
      call walktree(pnode%pnodekp, logunit, dmpunit, idebug)
    endif
  endif
  if (pnode%imother .ne. 4) then    ! we are not coming from i-1
    if (associated(pnode%pnodeim)) then
      call walktree(pnode%pnodeim, logunit, dmpunit, idebug)
    endif
  endif
  if (pnode%imother .ne. 5) then    ! we are not coming from j-1
    if (associated(pnode%pnodejm)) then
      call walktree(pnode%pnodejm, logunit, dmpunit, idebug)
    endif
  endif
  if (pnode%imother .ne. 6) then    ! we are not coming from k-1
    if (associated(pnode%pnodekm)) then
      call walktree(pnode%pnodekm, logunit, dmpunit, idebug)
    endif
  endif

  return
end subroutine walktree

!--------------------------------------------
! deletes tree
!--------------------------------------------

recursive subroutine deletetree(pnode, logunit, idebug)

  implicit none

  integer logunit, idebug

  type(treenode),pointer :: pnode

!----------------------------

  if (pnode%imother .ne. 1) then    ! we are not coming from i+1
    if (associated(pnode%pnodeip)) then
      call deletetree(pnode%pnodeip, logunit, idebug)
    endif
  endif
  if (pnode%imother .ne. 2) then    ! we are not coming from j+1
    if (associated(pnode%pnodejp)) then
      call deletetree(pnode%pnodejp, logunit, idebug)
    endif
  endif
  if (pnode%imother .ne. 3) then    ! we are not coming from k+1
    if (associated(pnode%pnodekp)) then
      call deletetree(pnode%pnodekp, logunit, idebug)
    endif
  endif
  if (pnode%imother .ne. 4) then    ! we are not coming from i-1
    if (associated(pnode%pnodeim)) then
      call deletetree(pnode%pnodeim, logunit, idebug)
    endif
  endif
  if (pnode%imother .ne. 5) then    ! we are not coming from j-1
    if (associated(pnode%pnodejm)) then
      call deletetree(pnode%pnodejm, logunit, idebug)
    endif
  endif
  if (pnode%imother .ne. 6) then    ! we are not coming from k-1
    if (associated(pnode%pnodekm)) then
      call deletetree(pnode%pnodekm, logunit, idebug)
    endif
  endif

  deallocate(pnode)

  return
end subroutine deletetree

!=================================================

end module treelib


